@@ -102,12 +102,13 @@ pathstart(FILE *cmd, int layer, int x, int y, u_char special, double oscale,
 	    if (wvia2 > wvia) wvia = wvia2;
          }
 
-         fprintf(cmd, "%s %g ( %g %g ) ", CIFLayer[layer],
-			invscale * (int)(oscale * wvia + 0.5),
-			invscale * x, invscale * y);
+         fprintf(cmd, "%s %ld ( %ld %ld ) ", CIFLayer[layer],
+			(long)(0.5 + invscale * oscale * wvia),
+			(long)(0.5 + invscale * x), (long)(0.5 + invscale * y));
       }
       else
-         fprintf(cmd, "%s ( %g %g ) ", CIFLayer[layer], invscale * x, invscale * y);
+         fprintf(cmd, "%s ( %ld %ld ) ", CIFLayer[layer],
+			(long)(0.5 + invscale * x), (long)(0.5 + invscale * y));
    }
    Pathon = 1;
 
@@ -145,14 +146,14 @@ pathto(FILE *cmd, int x, int y, int horizontal, int lastx, int lasty,
 
     fprintf(cmd, "( ");
     if (horizontal)
-	fprintf(cmd, "%g ", invscale * x);
+	fprintf(cmd, "%ld ", (long)(0.5 + invscale * x));
     else
 	fprintf(cmd, "* ");
 
     if (horizontal)
 	fprintf(cmd, "* ");
     else
-	fprintf(cmd, "%g ", invscale * y);
+	fprintf(cmd, "%ld ", (long)(0.5 + invscale * y));
 
     fprintf(cmd, ") ");
 
@@ -172,6 +173,7 @@ pathvia(FILE *cmd, int layer, int x, int y, int lastx, int lasty,
 {
     char *s;
     char checkersign = (gridx + gridy + layer) & 0x01;
+    NODEINFO lnode;
 
     if ((ViaPattern == VIA_PATTERN_NONE) || (ViaY[layer] == NULL))
 	s = ViaX[layer];
@@ -180,12 +182,26 @@ pathvia(FILE *cmd, int layer, int x, int y, int lastx, int lasty,
     else
 	s = (checkersign == 0) ?  ViaY[layer] : ViaX[layer];
 
+    /* If the position is a node, then the via type may be switched if	*/
+    /* there is a prohibition declared in the flags.			*/
+
+    if (layer < Pinlayers) {
+	if (((lnode = NODEIPTR(gridx, gridy, layer)) != NULL)
+			&& (lnode->nodesav != NULL)) {
+	    if ((lnode->flags & NI_NO_VIAX) && (s == ViaX[layer]))
+		s = ViaY[layer];
+	    if ((lnode->flags & NI_NO_VIAY) && (s == ViaY[layer]))
+		s = ViaX[layer];
+	}
+    }
+
     if (Pathon <= 0) {
        if (Pathon == -1)
 	  fprintf(cmd, "+ ROUTED ");
        else 
 	  fprintf(cmd, "\n  NEW ");
-       fprintf(cmd, "%s ( %g %g ) ", CIFLayer[layer], invscale * x, invscale * y);
+       fprintf(cmd, "%s ( %ld %ld ) ", CIFLayer[layer],
+		(long)(0.5 + invscale * x), (long)(0.5 + invscale * y));
     }
     else {
        // Normally the path will be manhattan and only one of
@@ -554,6 +570,11 @@ void print_nlnets( char *filename )
 /* is easier just to find any such instances and remove them by	*/
 /* eliminating one of the vias and adding a segment to connect	*/
 /* the route to the neighboring via.				*/
+/*								*/
+/* Note that the ensuing change in connectivity can violate	*/
+/* the route endpoints and thereby mess up the delay output	*/
+/* routine unless route_set_connections() is re-run on the	*/
+/* modified routes.						*/
 /*--------------------------------------------------------------*/
 
 static void cleanup_net(NET net)
@@ -562,10 +583,11 @@ static void cleanup_net(NET net)
    ROUTE rt, rt2;
    NODEINFO lnode;
    int lf, ll, lf2, ll2;
-   u_char fcheck, lcheck;
+   u_char fcheck, lcheck, needfix;
    u_char xcheckf, ycheckf, xcheckl, ycheckl; 
 
    lf = ll = lf2 = ll2 = -1;
+   needfix = FALSE;
 
    for (rt = net->routes; rt; rt = rt->next) {
       fcheck = lcheck = FALSE;
@@ -597,6 +619,13 @@ static void cleanup_net(NET net)
 			&& (lnode->nodesav != NULL))
 	    fcheck = FALSE;
       }
+
+      /* This could be done if vias are always too close when	*/
+      /* placed on adjacent tracks.  However, that ignores the	*/
+      /* problem of vias with offsets, and it ignores the fact	*/
+      /* that adjacent vias on the same net are always a	*/
+      /* redundancy. 						*/
+
       xcheckf = needblock[lf] & VIABLOCKX;
       ycheckf = needblock[lf] & VIABLOCKY;
       if (!xcheckf && !ycheckf) fcheck = FALSE;
@@ -624,6 +653,7 @@ static void cleanup_net(NET net)
 			&& (lnode->nodesav != NULL))
 	    lcheck = FALSE;
       }
+
       xcheckl = needblock[ll] & VIABLOCKX;
       ycheckl = needblock[ll] & VIABLOCKY;
       if (!xcheckl && !ycheckl) lcheck = FALSE;
@@ -642,6 +672,7 @@ static void cleanup_net(NET net)
 		  if ((seg->layer == lf) || ((seg->layer + 1) == lf)) {
 		     if (xcheckf && (seg->y1 == segf->y1) &&
 				(ABSDIFF(seg->x1, segf->x1) == 1)) {
+			needfix = TRUE;
 			if (seg->layer != segf->layer) {
 
 			   // Adjacent vias are different types.
@@ -672,6 +703,7 @@ static void cleanup_net(NET net)
 		     }
 		     else if (ycheckf && (seg->x1 == segf->x1) &&
 				(ABSDIFF(seg->y1, segf->y1) == 1)) {
+			needfix = TRUE;
 			if (seg->layer != segf->layer) {
 			   // Adjacent vias are different types.
 			   // Deal with it by creating a route between
@@ -706,6 +738,7 @@ static void cleanup_net(NET net)
 		  if ((seg->layer == ll) || ((seg->layer + 1) == ll)) {
 		     if (xcheckl && (seg->y1 == segl->y1) &&
 				(ABSDIFF(seg->x1, segl->x1) == 1)) {
+			needfix = TRUE;
 			if (seg->layer != segl->layer) {
 
 			   // Adjacent vias are different types.
@@ -736,6 +769,7 @@ static void cleanup_net(NET net)
 		     }
 		     else if (ycheckl && (seg->x1 == segl->x1) &&
 				(ABSDIFF(seg->y1, segl->y1) == 1)) {
+			needfix = TRUE;
 			if (seg->layer != segl->layer) {
 
 			   // Adjacent vias are different types.
@@ -770,6 +804,9 @@ static void cleanup_net(NET net)
 	 }
       }
    }
+   if (needfix == TRUE)
+      for (rt = net->routes; rt; rt = rt->next)
+	 route_set_connections(net, rt);
 }
 
 /*--------------------------------------------------------------*/
@@ -1143,11 +1180,13 @@ emit_routed_net(FILE *Cmd, NET net, u_char special, double oscale, int iscale)
 		  dir2 = OBSVAL(seg->next->x1, seg->next->y1, seg->next->layer) &
 					OFFSET_TAP;
 		  lnode2 = NODEIPTR(seg->next->x1, seg->next->y1, seg->next->layer);
-		  offset2 = lnode2->offset;
+		  if (lnode2 != NULL)
+		      offset2 = lnode2->offset;
 	       }
 	       else {
 		  lnode2 = NODEIPTR(seg->x2, seg->y2, seg->layer);
-		  offset2 = lnode2->offset;
+		  if (lnode2 != NULL)
+		      offset2 = lnode2->offset;
 	       }
 
 	       // Offset was calculated for vias;  plain metal routes
@@ -1226,8 +1265,18 @@ emit_routed_net(FILE *Cmd, NET net, u_char special, double oscale, int iscale)
 				" at (%d %d) to (%d %d)\n", x, y, x2, y2);
 		     }
 		     if (special == (u_char)0) {
-			pathstart(Cmd, seg->layer, x, y, special, oscale, invscale,
-				horizontal);
+			if (lastseg && (lastseg->segtype & ST_OFFSET_START) &&
+				((lastx != x) || (lasty != y))) {
+			   /* Add bend and connect to offset via */
+			   int vertical = (horizontal) ? FALSE : TRUE;
+			   pathstart(Cmd, seg->layer, lastx, lasty, special,
+				oscale, invscale, vertical);
+			   pathto(Cmd, x, y, vertical, lastx, lasty, invscale);
+			}
+			else {
+			   pathstart(Cmd, seg->layer, x, y, special, oscale,
+				invscale, horizontal);
+			}
 			lastx = x;
 			lasty = y;
 			lastlay = seg->layer;
